﻿using System;
using System.Net.Sockets;
using System.Threading;
using System.Windows.Forms;
using SDRSharp.Common;
using SDRSharp.Radio;
using System.Text;
using SDRSharp.Radio.PortAudio;

namespace SDRSharp.FrontEnds.SpyServer
{
    public unsafe class SpyClient : IDisposable
    {
        private const int BufferSize = 64 * 1024;
        private const int DefaultDisplayPixels = 2000;
        private const int DefaultFFTRange = 127;

        enum ParserPhase
        {
            AcquiringHeader,
            ReadingData
        }

        private Socket _s;
        private Thread _receiveThread;
        private bool _terminated;
        private bool _streaming;
        private bool _useCompression;
        private StreamingMode _streamingMode = StreamingMode.STREAM_MODE_FFT_IQ;
        private bool _optimizePropertyChanges = true;
        private uint _channelCenterFrequency;
        private uint _displayCenterFrequency;
        private uint _deviceCenterFrequency;
        private int _displayDecimationStageCount;
        private int _channelDecimationStageCount;
        private uint _minimumTunableFrequency;
        private uint _maximumTunableFrequency;
        private int _gain;
        private int _fftOffset;
        private int _fftRange = DefaultFFTRange;
        private int _displayPixels = DefaultDisplayPixels;
        private int _messageSize;
        private uint _lastSequenceNumber = uint.MaxValue;
        private uint _droppedBuffers;
        private long _down_stream_bytes;
        private bool _gotDeviceInfo;
        private bool _gotSyncInfo;
        private bool _canControl;
        private Exception _error;
        private DeviceInfo _deviceInfo;
        private ParserPhase _parserPhase;
        private int _parserPosition;
        private byte[] _headerData = new byte[sizeof(MessageHeader)];
        private MessageHeader _header;
        private UnsafeBuffer _bodyBuffer;
        private UnsafeBuffer _uncompressedBuffer;
        private UnsafeBuffer _messageBuffer;
        private UnsafeBuffer _iqBuffer;

        public event EventHandler Connected;
        public event EventHandler Disconnected;
        public event EventHandler Synchronized;
        public event SamplesAvailableDelegate<ComplexSamplesEventArgs> SamplesAvailable;
        public event SamplesAvailableDelegate<ByteSamplesEventArgs> FFTAvailable;

        public void Dispose()
        {
            Disconnect();
        }

        public void Connect(string host, int port)
        {
            if (_receiveThread != null)
            {
                return;
            }

            _s = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            _s.NoDelay = true;
            _s.Connect(host, port);
            _s.Blocking = false;

            SayHello();

            Cleanup();

            _error = null;
            _terminated = false;
            _receiveThread = new Thread(new ThreadStart(ThreadProc));
            _receiveThread.Name = "Spy Server Receive Thread";
            _receiveThread.Start();

            for (int i = 0; i < 2000 && _error == null; i++)
            {
                if (_gotDeviceInfo)
                {
                    if (_deviceInfo.DeviceType == 0)
                    {
                        _error = new ApplicationException("Server is up but no device is available.");
                        break;
                    }

                    if (_gotSyncInfo)
                    {
                        OnConnect();
                        return;
                    }
                }
                Thread.Sleep(1);
                Application.DoEvents();
            }

            Disconnect();

            if (_error != null)
            {
                var e = _error;
                _error = null;
                throw e;
            }

            throw new ApplicationException("Server didn't send the device capability and synchronization info.");
        }

        public void Disconnect()
        {
            _terminated = true;

            if (_s != null)
            {
                _s.Close();
                _s = null;
            }

            var thread = _receiveThread;
            if (thread != null)
            {
                thread.Join();
                _receiveThread = null;
            }

            Cleanup();
        }

        private void OnConnect()
        {
            SetSetting(SettingType.SETTING_STREAMING_MODE, (uint) _streamingMode);

            SetSetting(SettingType.SETTING_FFT_DISPLAY_PIXELS, (uint) _displayPixels);
            SetSetting(SettingType.SETTING_FFT_DB_OFFSET, (uint) _fftOffset);
            SetSetting(SettingType.SETTING_FFT_DB_RANGE, (uint) _fftRange);

            UpdateFFTFormat();
            UpdateIQFormat();
        }

        public void StartStreaming()
        {
            if (!_streaming)
            {
                _streaming = true;
                _down_stream_bytes = 0;
                SetStreamState();
            }
        }

        public void StopStreaming()
        {
            if (_streaming)
            {
                _streaming = false;
                _down_stream_bytes = 0;
                SetStreamState();
            }
        }

        private void Cleanup()
        {
            _deviceInfo.DeviceType = 0;
            _deviceInfo.DeviceSerial = 0;
            _deviceInfo.DecimationStageCount = 0;
            _deviceInfo.GainStageCount = 0;
            _deviceInfo.MaximumSampleRate = 0;
            _deviceInfo.MaximumBandwidth = 0;
            _deviceInfo.MaximumGainIndex = 0;
            _deviceInfo.MinimumFrequency = 0;
            _deviceInfo.MaximumFrequency = 0;
            _deviceInfo.Resolution = 0;

            _gain = 0;
            _displayCenterFrequency = 0;
            _deviceCenterFrequency = 0;
            _displayDecimationStageCount = 0;
            _channelDecimationStageCount = 0;
            _minimumTunableFrequency = 0;
            _maximumTunableFrequency = 0;
            _canControl = false;
            _gotDeviceInfo = false;
            _gotSyncInfo = false;

            _lastSequenceNumber = uint.MaxValue;
            _droppedBuffers = 0;
            _down_stream_bytes = 0;

            _parserPhase = ParserPhase.AcquiringHeader;
            _parserPosition = 0;

            _streaming = false;
            _terminated = true;
        }

        #region Properties

        public bool UseCompression
        {
            get
            {
                return _useCompression;
            }
            set
            {
                if (_useCompression != value)
                {
                    _useCompression = value;
                    UpdateFFTFormat();
                    UpdateIQFormat();
                }
            }
        }

        public bool OptimizePropertyChanges
        {
            get { return _optimizePropertyChanges; }
            set { _optimizePropertyChanges = value; }
        }

        public bool IsConnected
        {
            get
            {
                return !_terminated && _s != null && _s.Connected;
            }
        }

        public bool IsSynchronized
        {
            get
            {
                return _gotDeviceInfo && _gotSyncInfo;
            }
        }

        public string DeviceName
        {
            get
            {
                return Constants.GetDeviceName(_deviceInfo.DeviceType);
            }
        }

        public uint DeviceSerial
        {
            get
            {
               return _deviceInfo.DeviceSerial;
            }
        }

        public uint MaximumDecimationStageCount
        {
            get
            {
                return _deviceInfo.DecimationStageCount;
            }
        }

        public uint MaximumBandwidth
        {
            get
            {
                return _deviceInfo.MaximumBandwidth;
            }
        }

        public uint MaximumSampleRate
        {
            get
            {
                return _deviceInfo.MaximumSampleRate;
            }
        }

        public uint MaximumGainIndex
        {
            get
            {
                return _deviceInfo.MaximumGainIndex;
            }
        }

        public double ChannelSamplerate
        {
            get
            {
                return _deviceInfo.MaximumSampleRate / (double) (1 << _channelDecimationStageCount);
            }
        }

        public int DisplayBandwidth
        {
            get
            {
                return (int) (_deviceInfo.MaximumBandwidth / (double) (1 << _displayDecimationStageCount));
            }
        }

        public int ChannelBandwidth
        {
            get
            {
                return (int) (_deviceInfo.MaximumBandwidth / (double) (1 << _channelDecimationStageCount));
            }
        }

        public bool CanControl
        {
            get
            {
                return _canControl;
            }
        }

        public uint MaximumTunableFrequency
        {
            get
            {
                return _maximumTunableFrequency;
            }
        }

        public uint MinimumTunableFrequency
        {
            get
            {
                return _minimumTunableFrequency;
            }
        }

        public uint DeviceCenterFrequency
        {
            get
            {
                return _deviceCenterFrequency;
            }
        }

        public StreamingMode StreamingMode
        {
            get
            {
                return _streamingMode;
            }
            set
            {
                if (_streamingMode != value || !_optimizePropertyChanges)
                {
                    _streamingMode = value;
                    SetSetting(SettingType.SETTING_STREAMING_MODE, (uint) _streamingMode);
                }
            }
        }

        public uint DisplayCenterFrequency
        {
            get
            {
                return _displayCenterFrequency;
            }
            set
            {
                if (_canControl)
                {
                    _deviceCenterFrequency = value;
                }
                if (_displayCenterFrequency != value || !_optimizePropertyChanges)
                {
                    _displayCenterFrequency = value;
                    SetSetting(SettingType.SETTING_FFT_FREQUENCY, _displayCenterFrequency);
                }
            }
        }

        public uint ChannelCenterFrequency
        {
            get
            {
                return _channelCenterFrequency;
            }
            set
            {
                if (_channelCenterFrequency != value || !_optimizePropertyChanges)
                {
                    _channelCenterFrequency = value;
                    SetSetting(SettingType.SETTING_IQ_FREQUENCY, _channelCenterFrequency);
                }
            }
        }

        public int Gain
        {
            get
            {
                return _gain;
            }
            set
            {
                if (_gain != value || !_optimizePropertyChanges)
                {
                    _gain = value;
                    SetSetting(SettingType.SETTING_GAIN, (uint) _gain);
                }
            }
        }

        public int DisplayDecimationStageCount
        {
            get
            {
                return _displayDecimationStageCount;
            }
            set
            {
                if (_displayDecimationStageCount != value || !_optimizePropertyChanges)
                {
                    _displayDecimationStageCount = value;
                    SetSetting(SettingType.SETTING_FFT_DECIMATION, (uint) _displayDecimationStageCount);
                }
            }
        }

        public int ChannelDecimationStageCount
        {
            get
            {
                return _channelDecimationStageCount;
            }
            set
            {
                if (_channelDecimationStageCount != value || !_optimizePropertyChanges)
                {
                    _channelDecimationStageCount = value;
                    SetSetting(SettingType.SETTING_IQ_DECIMATION, (uint) _channelDecimationStageCount);
                    UpdateIQFormat();
                }
            }
        }

        public int FFTRange
        {
            get
            {
                return _fftRange;
            }
            set
            {
                if (_fftRange != value || !_optimizePropertyChanges)
                {
                    _fftRange = value;
                    SetSetting(SettingType.SETTING_FFT_DB_RANGE, (uint) _fftRange);
                }
            }
        }

        public int FFTOffset
        {
            get
            {
                return _fftOffset;
            }
            set
            {
                if (_fftOffset != value || !_optimizePropertyChanges)
                {
                    _fftOffset = value;
                    SetSetting(SettingType.SETTING_FFT_DB_OFFSET, (uint) _fftOffset);
                }
            }
        }

        public int DisplayPixels
        {
            get
            {
                return _displayPixels;
            }
            set
            {
                if (_displayPixels != value || !_optimizePropertyChanges)
                {
                    _displayPixels = value;
                    SetSetting(SettingType.SETTING_FFT_DISPLAY_PIXELS, (uint) _displayPixels);
                }
            }
        }

        #endregion

        #region Client Commands

        private void UpdateFFTFormat()
        {
            var format = StreamFormat.STREAM_FORMAT_UINT8;

            if (_useCompression)
            {
                format |= StreamFormat.STREAM_FORMAT_COMPRESSED;
            }

            SetSetting(SettingType.SETTING_FFT_FORMAT, (uint) format);
        }

        private void UpdateIQFormat()
        {
            var requiredBits = _deviceInfo.Resolution + (int) Math.Ceiling(_channelDecimationStageCount * 0.5) - 1;

            StreamFormat format;

            if (requiredBits <= 8)
            {
                format = StreamFormat.STREAM_FORMAT_UINT8;
            }
            else if (requiredBits <= 16)
            {
                format = StreamFormat.STREAM_FORMAT_INT16;
            }
            else if (requiredBits <= 24)
            {
                format = StreamFormat.STREAM_FORMAT_INT24;
            }
            else
            {
                format = StreamFormat.STREAM_FORMAT_FLOAT;
            }

            if (_useCompression)
            {
                format |= StreamFormat.STREAM_FORMAT_COMPRESSED;
            }

            SetSetting(SettingType.SETTING_IQ_FORMAT, (uint) format);
        }

        private bool SetStreamState()
        {
            return SetSetting(SettingType.SETTING_STREAMING_ENABLED, (uint) (_streaming ? 1 : 0));
        }

        private bool SetSetting(SettingType setting, params uint[] args)
        {
            byte[] argBytes;
            if (args != null && args.Length > 0)
            {
                argBytes = new byte[sizeof(SettingType) + args.Length * sizeof(uint)];

                var settingBytes = (byte*) &setting;
                for (int i = 0; i < sizeof(SettingType); i++)
                {
                    argBytes[i] = settingBytes[i];
                }

                Buffer.BlockCopy(args, 0, argBytes, sizeof(SettingType), args.Length * sizeof(uint));
            }
            else
            {
                argBytes = null;
            }

            return SendCommand(CommandType.CMD_SET_SETTING, argBytes);
        }

        private bool SayHello()
        {
            var protocolVersionBytes = BitConverter.GetBytes(Constants.SPYSERVER_PROTOCOL_VERSION);

            var softwareID = string.Format("SDR# v{0} on {1}", System.Reflection.Assembly.GetEntryAssembly().GetName().Version, Environment.OSVersion);
            var softwareVersionBytes = Encoding.ASCII.GetBytes(softwareID);

            var args = new byte[protocolVersionBytes.Length + softwareVersionBytes.Length];

            Buffer.BlockCopy(protocolVersionBytes, 0, args, 0, protocolVersionBytes.Length);
            Buffer.BlockCopy(softwareVersionBytes, 0, args, protocolVersionBytes.Length, softwareVersionBytes.Length);

            return SendCommand(CommandType.CMD_HELLO, args);
        }

        private bool SendCommand(CommandType cmd, byte[] args)
        {
            if (_s == null)
            {
                return false;
            }

            var headerLen = sizeof(CommandHeader);
            var argLen = args == null ? 0 : args.Length;

            var buffer = new byte[headerLen + argLen];

            CommandHeader header;
            header.CommandType = cmd;
            header.BodySize = (ushort) argLen;

            var headerBytes = (byte *) &header;
            for (int i = 0; i < sizeof(CommandHeader); i++)
            {
                buffer[i] = headerBytes[i];
            }

            if (args != null)
            {
                fixed (byte* ptr = buffer)
                {
                    var pargs = ptr + headerLen;
                    for (int i = 0; i < args.Length; i++)
                    {
                        pargs[i] = args[i];
                    }
                }
            }

            try
            {
                _s.Send(buffer);
            }
            catch
            {
                return false;
            }
            return true;
        }

        #endregion

        #region Message Handling

        private void ThreadProc()
        {
            Connected?.Invoke(this, EventArgs.Empty);

            _parserPhase = ParserPhase.AcquiringHeader;
            _parserPosition = 0;

            var buffer = new byte[BufferSize];
            fixed (byte* ptr = buffer)
            {
                try
                {
                    while (!_terminated)
                    {
                        var s = _s;
                        if (s != null && s.Poll(1000000, SelectMode.SelectRead))
                        {
                            if (_terminated)
                            {
                                break;
                            }
                            var len = s.Receive(buffer, 0, buffer.Length, SocketFlags.None);
                            if (len > 0)
                            {
                                ParseMessage(ptr, len);
                            }
                            else
                            {
                                throw new ApplicationException("Device got disconnected");
                            }
                        }
                    }
                }
                catch (Exception e)
                {
                    _error = e;
                }
            }

            if (_bodyBuffer != null)
            {
                _bodyBuffer.Dispose();
                _bodyBuffer = null;
            }

            if (_uncompressedBuffer != null)
            {
                _uncompressedBuffer.Dispose();
                _uncompressedBuffer = null;
            }

            _messageBuffer = null;

            Cleanup();
            _receiveThread = null;
            Disconnected?.Invoke(this, EventArgs.Empty);
        }
        private int ParseHeader(byte* buffer, int length)
        {
            var consumed = 0;

            fixed (byte* headerPtr = _headerData)
            {
                while (length > 0)
                {
                    int to_write = Math.Min(sizeof(MessageHeader) - _parserPosition, length);
                    Utils.Memcpy(headerPtr + _parserPosition, buffer, to_write);
                    length -= to_write;
                    buffer += to_write;
                    _parserPosition += to_write;
                    consumed += to_write;

                    if (_parserPosition == sizeof(MessageHeader))
                    {
                        _parserPosition = 0;

                        MessageHeader header;

                        Utils.Memcpy(&header, headerPtr, sizeof(MessageHeader));

                        _header = header;

                        if (header.BodySize > 0)
                        {
                            _parserPhase = ParserPhase.ReadingData;
                        }

                        return consumed;
                    }
                }
            }

            return consumed;
        }

        private int ParseBody(byte* buffer, int length)
        {
            var consumed = 0;

            var bodyPtr = (byte *) _bodyBuffer;
            while (length > 0)
            {
                int to_write = Math.Min((int) _header.BodySize - _parserPosition, length);
                Utils.Memcpy(bodyPtr + _parserPosition, buffer, to_write);
                length -= to_write;
                buffer += to_write;
                _parserPosition += to_write;
                consumed += to_write;

                if (_parserPosition == _header.BodySize)
                {
                    _parserPosition = 0;
                    _parserPhase = ParserPhase.AcquiringHeader;

                    return consumed;
                }
            }

            return consumed;
        }

        public long GetDownstreamBytes()
        {
            return Interlocked.Exchange(ref _down_stream_bytes, 0);
        }

        private void ParseMessage(byte *buffer, int len)
        {
            Interlocked.Add(ref _down_stream_bytes, len);

            int consumed;

            while (len > 0 && !_terminated)
            {
                if (_parserPhase == ParserPhase.AcquiringHeader)
                {
                    while (_parserPhase == ParserPhase.AcquiringHeader && len > 0)
                    {
                        consumed = ParseHeader(buffer, len);
                        buffer += consumed;
                        len -= consumed;
                    }

                    if (_parserPhase == ParserPhase.ReadingData)
                    {
                        var client_major = (byte) (Constants.SPYSERVER_PROTOCOL_VERSION >> 24);
                        var client_minor = (byte) ((Constants.SPYSERVER_PROTOCOL_VERSION >> 16) & 0xff);
                        var client_build = (ushort) (Constants.SPYSERVER_PROTOCOL_VERSION & 0xffff);
                        
                        var server_major = (byte) (_header.ProtocolID >> 24);
                        var server_minor = (byte) ((_header.ProtocolID >> 16) & 0xff);
                        var server_build = (ushort) (_header.ProtocolID & 0xffff);

                        if (client_major != server_major || client_minor != server_minor)
                        {
                            var message = string.Format("Server is running an unsupported protocol version.\r\nExpected {0}.{1}.* but got {3}.{4}.{5}.",
                                client_major, client_minor, client_build, server_major, server_minor, server_build);
                            throw new ApplicationException(message);
                        }

                        if (_header.BodySize > Constants.SPYSERVER_MAX_MESSAGE_BODY_SIZE)
                        {
                            throw new ApplicationException("The server is probably buggy");
                        }

                        if (_bodyBuffer == null || _bodyBuffer.Length < _header.BodySize)
                        {
                            if (_bodyBuffer != null)
                            {
                                _bodyBuffer.Dispose();
                            }
                            _bodyBuffer = UnsafeBuffer.Create((int) _header.BodySize);
                        }
                    }
                }

                if (_parserPhase == ParserPhase.ReadingData)
                {
                    consumed = ParseBody(buffer, len);
                    buffer += consumed;
                    len -= consumed;

                    if (_parserPhase == ParserPhase.AcquiringHeader)
                    {
                        if (_header.StreamType == StreamType.STREAM_TYPE_IQ)
                        {
                            var gap = _header.SequenceNumber - _lastSequenceNumber - 1;
                            _lastSequenceNumber = _header.SequenceNumber;
                            _droppedBuffers += gap;
                        }

                        HandleNewMessage();
                    }
                }
            }
        }

        private void HandleNewMessage()
        {
            if (_terminated)
            {
                return;
            }

            var isCompressed = (_header.MessageType & MessageType.MSG_TYPE_COMPRESSED) != 0;

            if (isCompressed)
            {
                if (_header.BodySize < sizeof(CompressionHeader))
                {
                    throw new ApplicationException("Corrupted message");
                }

                CompressionHeader ch;
                Utils.Memcpy(&ch, _bodyBuffer, sizeof(CompressionHeader));

                if (ch.UncompressedSize > Constants.SPYSERVER_MAX_MESSAGE_BODY_SIZE)
                {
                    throw new ApplicationException("Compressed message exceeds the maximum uncompressed size");
                }

                _messageSize = (int) ch.UncompressedSize;

                if (_uncompressedBuffer == null || _uncompressedBuffer.Length < _messageSize)
                {
                    if (_uncompressedBuffer != null)
                    {
                        _uncompressedBuffer.Dispose();
                    }

                    _uncompressedBuffer = UnsafeBuffer.Create(_messageSize);
                }

                _messageBuffer = _uncompressedBuffer;

                switch (ch.CompressionType)
                {
                    case CompressionType.COMPRESSION_TYPE_ARITHMETIC:
                        UncompressArithmeticBlob();
                        break;
                }
            }
            else
            {
                _messageBuffer = _bodyBuffer;
                _messageSize = (int) _header.BodySize;
            }

            var messageType = (MessageType) ((uint) _header.MessageType & Constants.SPYSERVER_MESSAGE_TYPE_MASK);

            switch (messageType)
            {
                case MessageType.MSG_TYPE_DEVICE_INFO:
                    ProcessDeviceInfo();
                    break;

                case MessageType.MSG_TYPE_CLIENT_SYNC:
                    ProcessClientSync();
                    break;

                case MessageType.MSG_TYPE_UINT8_IQ:
                    ProcessUInt8Samples();
                    break;

                case MessageType.MSG_TYPE_INT16_IQ:
                    ProcessInt16Samples();
                    break;

                case MessageType.MSG_TYPE_INT24_IQ:
                    ProcessInt24Samples();
                    break;

                case MessageType.MSG_TYPE_FLOAT_IQ:
                    ProcessFloatSamples();
                    break;

                case MessageType.MSG_TYPE_UINT8_FFT:
                    ProcessUInt8FFT();
                    break;
            }
        }

        private void UncompressArithmeticBlob()
        {
            var blob = (byte *) _bodyBuffer + sizeof(CompressionHeader);
            var blobSize = _header.BodySize - sizeof(CompressionHeader);

            var dest = (byte *) _messageBuffer;
            var destSize = _messageSize;

            // Todo: Check the model and Uncompress
        }

        private void ProcessDeviceInfo()
        {
            DeviceInfo info;
            Utils.Memcpy(&info, _messageBuffer, Math.Min(_messageSize, sizeof(DeviceInfo)));

            _deviceInfo = info;

            if (_deviceInfo.Resolution == 0)
            {
                // Fallback for old servers.
                switch (_deviceInfo.DeviceType)
                {
                    case DeviceType.DEVICE_AIRSPY_HF:
                        _deviceInfo.Resolution = 16;
                        break;

                    case DeviceType.DEVICE_AIRSPY_ONE:
                        _deviceInfo.Resolution = 12;
                        break;

                    case DeviceType.DEVICE_RTLSDR:
                        _deviceInfo.Resolution = 8;
                        break;
                }
            }

            _minimumTunableFrequency = _deviceInfo.MinimumFrequency;
            _maximumTunableFrequency = _deviceInfo.MaximumFrequency;

            _gotDeviceInfo = true;
        }

        private void ProcessClientSync()
        {
            ClientSync sync;
            Utils.Memcpy(&sync, _messageBuffer, Math.Min(_messageSize, sizeof(ClientSync)));

            _canControl = sync.CanControl != 0;
            _gain = (int) sync.Gain;
            _deviceCenterFrequency = sync.DeviceCenterFrequency;
            _channelCenterFrequency = sync.IQCenterFrequency;
            _displayCenterFrequency = sync.FFTCenterFrequency;

            switch (_streamingMode)
            {
                case StreamingMode.STREAM_MODE_FFT_ONLY:
                case StreamingMode.STREAM_MODE_FFT_IQ:
                    _minimumTunableFrequency = sync.MinimumFFTCenterFrequency;
                    _maximumTunableFrequency = sync.MaximumFFTCenterFrequency;
                    break;

                case StreamingMode.STREAM_MODE_IQ_ONLY:
                    _minimumTunableFrequency = sync.MinimumIQCenterFrequency;
                    _maximumTunableFrequency = sync.MaximumIQCenterFrequency;
                    break;
            }

            Synchronized?.Invoke(this, EventArgs.Empty);

            _gotSyncInfo = true;
        }

        #endregion

        #region Process IQ Samples

        private void ProcessUInt8Samples()
        {
            const float scale = 1.0f / 128;

            var sampleCount = _messageSize / 2;

            if (_iqBuffer == null || _iqBuffer.Length != sampleCount)
            {
                _iqBuffer = UnsafeBuffer.Create(sampleCount, sizeof(Complex));
            }

            var input_samples = (byte*) _messageBuffer;
            var output_samples = (Complex*) _iqBuffer;

            for (var i = 0; i < sampleCount; i++)
            {
                output_samples[i].Real = (*input_samples++ - 128) * scale;
                output_samples[i].Imag = (*input_samples++ - 128) * scale;
            }

            PushIQData(output_samples, sampleCount);
        }

        private void ProcessInt16Samples()
        {
            const float scale = 1.0f / 32768;

            var sampleCount = _messageSize / 4;

            if (_iqBuffer == null || _iqBuffer.Length != sampleCount)
            {
                _iqBuffer = UnsafeBuffer.Create(sampleCount, sizeof(Complex));
            }

            var input_samples = (short*) _messageBuffer;
            var output_samples = (Complex*) _iqBuffer;

            for (var i = 0; i < sampleCount; i++)
            {
                output_samples[i].Real = *input_samples++ * scale;
                output_samples[i].Imag = *input_samples++ * scale;
            }

            PushIQData(output_samples, sampleCount);
        }

        private void ProcessInt24Samples()
        {
            const float scale = 1.0f / 8388607.0f;

            var sampleCount = _messageSize / 6;

            if (_iqBuffer == null || _iqBuffer.Length != sampleCount)
            {
                _iqBuffer = UnsafeBuffer.Create(sampleCount, sizeof(Complex));
            }

            var input_samples = (Int24*) _messageBuffer;
            var output_samples = (Complex*) _iqBuffer;

            for (var i = 0; i < sampleCount; i++)
            {
                output_samples[i].Real = *input_samples++ * scale;
                output_samples[i].Imag = *input_samples++ * scale;
            }

            PushIQData(output_samples, sampleCount);
        }

        private void ProcessFloatSamples()
        {
            var sampleCount = _messageSize / 8;            
            var output_samples = (Complex*) _messageBuffer;
            PushIQData(output_samples, sampleCount);
        }

        private void PushIQData(Complex* samples, int count)
        {
            var args = new ComplexSamplesEventArgs()
            {
                Buffer = samples,
                Length = count,
                DroppedSamples = _droppedBuffers * (uint) count
            };

            _droppedBuffers = 0;

            SamplesAvailable?.Invoke(this, args);
        }

        #endregion

        #region Process FFT

        private void ProcessUInt8FFT()
        {
            var args = new ByteSamplesEventArgs()
            {
                Buffer = (byte*) _messageBuffer,
                Length =  _messageSize
            };

            FFTAvailable?.Invoke(this, args);
        }

        #endregion
    }
}
